/**
 * @license
 * Copyright (c) 2016 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at http://polymer.github.io/PATENTS.txt
 */
"use strict";
const hydrolysis_1 = require('hydrolysis');
const path = require('path');
const stream_1 = require('stream');
const url_1 = require('url');
const logging = require('plylog');
const dom5_1 = require('dom5');
const minimatchAll = require('minimatch-all');
let logger = logging.getLogger('cli.build.analyzer');
class StreamAnalyzer extends stream_1.Transform {
    constructor(root, entrypoint, shell, fragments) {
        super({ objectMode: true });
        this.files = new Map();
        this.root = root;
        this.entrypoint = entrypoint;
        this.shell = shell;
        this.fragments = fragments;
        this.allFragments = [];
        // It's important that shell is first for document-ordering of imports
        if (shell) {
            this.allFragments.push(shell);
        }
        if (entrypoint && !shell && fragments.length === 0) {
            this.allFragments.push(entrypoint);
        }
        if (fragments) {
            this.allFragments = this.allFragments.concat(fragments);
        }
        this.resolver = new StreamResolver(this);
        this.loader = new hydrolysis_1.Loader();
        this.loader.addResolver(this.resolver);
        this.analyzer = new hydrolysis_1.Analyzer(false, this.loader);
        this.analyze = new Promise((resolve, reject) => {
            this._analyzeResolve = resolve;
        });
    }
    _transform(file, encoding, callback) {
        // store the file for access by the resolver
        this.files.set(file.path, file);
        // If this is the entrypoint, hold on to the file, so that it's fully
        // analyzed by the time down-stream transforms see it.
        if (this.isFragment(file)) {
            callback(null, null);
        }
        else {
            callback(null, file);
        }
    }
    _flush(done) {
        this._getDepsToEntrypointIndex().then((depsIndex) => {
            // push held back files
            for (let entrypoint of this.allFragments) {
                let file = this.files.get(entrypoint);
                this.push(file);
            }
            this._analyzeResolve(depsIndex);
            done();
        });
    }
    /**
     * A side-channel to add files to the resolver that did not come throgh the
     * stream transformation. This is for generated files, like
     * shared-bundle.html. This should probably be refactored so that the files
     * can be injected into the stream.
     */
    addFile(file) {
        this.files.set(file.path, file);
    }
    isFragment(file) {
        return this.allFragments.indexOf(file.path) !== -1;
    }
    _getDepsToEntrypointIndex() {
        // TODO: tsc is being really weird here...
        let depsPromises = this.allFragments.map((e) => this._getDependencies(e));
        return Promise.all(depsPromises).then((value) => {
            // tsc was giving a spurious error with `allDeps` as the parameter
            let allDeps = value;
            // An index of dependency -> fragments that depend on it
            let depsToFragments = new Map();
            // An index of fragments -> dependencies
            let fragmentToDeps = new Map();
            let fragmentToFullDeps = new Map();
            console.assert(this.allFragments.length === allDeps.length);
            for (let i = 0; i < allDeps.length; i++) {
                let fragment = this.allFragments[i];
                let deps = allDeps[i];
                console.assert(deps != null, `deps is null for ${fragment}`);
                fragmentToDeps.set(fragment, deps.imports);
                fragmentToFullDeps.set(fragment, deps);
                for (let dep of deps.imports) {
                    let entrypointList;
                    if (!depsToFragments.has(dep)) {
                        entrypointList = [];
                        depsToFragments.set(dep, entrypointList);
                    }
                    else {
                        entrypointList = depsToFragments.get(dep);
                    }
                    entrypointList.push(fragment);
                }
            }
            return {
                depsToFragments: depsToFragments,
                fragmentToDeps: fragmentToDeps,
                fragmentToFullDeps: fragmentToFullDeps,
            };
        });
    }
    /**
     * Attempts to retreive document-order transitive dependencies for `url`.
     */
    _getDependencies(url) {
        let documents = this.analyzer.parsedDocuments;
        let dir = path.dirname(url);
        return this.analyzer.metadataTree(url)
            .then((tree) => this._getDependenciesFromDescriptor(tree, dir));
    }
    _getDependenciesFromDescriptor(descriptor, dir) {
        let allHtmlDeps = [];
        let allScriptDeps = new Set();
        let allStyleDeps = new Set();
        let deps = this._collectScriptsAndStyles(descriptor);
        deps.scripts.forEach((s) => allScriptDeps.add(path.resolve(dir, s)));
        deps.styles.forEach((s) => allStyleDeps.add(path.resolve(dir, s)));
        if (descriptor.imports) {
            let queue = descriptor.imports.slice();
            while (queue.length > 0) {
                let next = queue.shift();
                if (!next.href) {
                    continue;
                }
                allHtmlDeps.push(next.href);
                let childDeps = this._getDependenciesFromDescriptor(next, path.dirname(next.href));
                allHtmlDeps = allHtmlDeps.concat(childDeps.imports);
                childDeps.scripts.forEach((s) => allScriptDeps.add(s));
                childDeps.styles.forEach((s) => allStyleDeps.add(s));
            }
        }
        return {
            scripts: Array.from(allScriptDeps),
            styles: Array.from(allStyleDeps),
            imports: allHtmlDeps,
        };
    }
    _collectScriptsAndStyles(tree) {
        let scripts = [];
        let styles = [];
        tree.html.script.forEach((script) => {
            if (script['__hydrolysisInlined']) {
                scripts.push(script['__hydrolysisInlined']);
            }
        });
        tree.html.style.forEach((style) => {
            let href = dom5_1.getAttribute(style, 'href');
            if (href) {
                styles.push(href);
            }
        });
        return {
            scripts: scripts,
            styles: styles
        };
    }
}
exports.StreamAnalyzer = StreamAnalyzer;
class StreamResolver {
    constructor(analyzer) {
        this.analyzer = analyzer;
    }
    accept(url, deferred) {
        var parsed = url_1.parse(url);
        var filepath;
        if (!parsed.hostname) {
            filepath = parsed.pathname;
        }
        // this.analyzer.requestedUrls.add(local);
        if (filepath) {
            // un-escape HTML escapes
            filepath = decodeURIComponent(filepath);
            // If the file path is not already under root, such as /bower_components/...,
            // prefix it with root
            if (!filepath.startsWith(this.analyzer.root)) {
                filepath = path.join(this.analyzer.root, filepath);
            }
            let file = this.analyzer.files.get(filepath);
            if (file) {
                deferred.resolve(file.contents.toString());
            }
            else {
                logger.info('No file found for', filepath);
            }
            return true;
        }
        return false;
    }
}