/**
 * @license
 * Copyright (c) 2016 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at http://polymer.github.io/PATENTS.txt
 */
"use strict";
const fs = require('fs');
const GitHubApi = require('github');
const path = require('path');
const gunzip = require('gunzip-maybe');
const request = require('request');
const tar = require('tar-fs');
class Github {
    constructor(opts) {
        this._token = opts.githubToken || Github.tokenFromFile('token');
        this._owner = opts.owner;
        this._repo = opts.repo;
        this._github = opts.githubApi || new GitHubApi({
            version: '3.0.0',
            protocol: 'https',
        });
        if (this._token != null) {
            this._github.authenticate({
                type: 'oauth',
                token: this._token,
            });
        }
        this._request = opts.requestApi || request;
    }
    static tokenFromFile(filename) {
        try {
            return fs.readFileSync(filename, 'utf8').trim();
        }
        catch (e) {
            return null;
        }
    }
    extractLatestRelease(outDir) {
        return new Promise((resolve, reject) => {
            let tarPipe = tar.extract(outDir, {
                ignore: (_, header) => {
                    let splitPath = header.name.split(path.sep);
                    // ignore the top directory in the tarfile to unpack directly to
                    // the cwd
                    return splitPath.length < 2 || splitPath[1] === '';
                },
                map: (header) => {
                    let unprefixed = header.name.split(path.sep).slice(1).join(path.sep).trim();
                    // the ./ is needed to unpack top-level files in the tar, otherwise
                    // they're just not written
                    header.name = './' + unprefixed;
                    return header;
                },
            });
            this.getLatestRelease().then((release) => {
                let tarballUrl = release.tarball_url;
                this._request({
                    url: tarballUrl,
                    headers: {
                        'User-Agent': 'request',
                        'Authorization': `token ${this._token}`
                    }
                })
                    .on('end', () => {
                    resolve();
                })
                    .pipe(gunzip())
                    .pipe(tarPipe)
                    .on('error', (error) => {
                    throw error;
                });
            })
                .catch((error) => {
                reject(error);
            });
        });
    }
    getLatestRelease() {
        return new Promise((resolve, reject) => {
            this._github.releases.listReleases({
                owner: this._owner,
                repo: this._repo
            }, (err, result) => {
                if (err) {
                    reject(err);
                }
                else {
                    if (result.length === 0) {
                        reject('no releases');
                    }
                    else {
                        resolve(result[0]);
                    }
                }
            });
        });
    }
}
exports.Github = Github;