/**
 * @license
 * Copyright (c) 2016 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at http://polymer.github.io/PATENTS.txt
 */
"use strict";
const commandLineCommands = require('command-line-commands');
const logging = require('plylog');
const args_1 = require('./args');
const build_1 = require('./commands/build');
const help_1 = require('./commands/help');
const init_1 = require('./commands/init');
const lint_1 = require('./commands/lint');
const serve_1 = require('./commands/serve');
const test_1 = require('./commands/test');
const project_config_1 = require('./project-config');
const logger = logging.getLogger('cli.main');
process.on('uncaughtException', (error) => {
    logger.error(`Uncaught exception: ${error}`);
    if (error.stack)
        logger.error(error.stack);
});
process.on('unhandledRejection', (error) => {
    logger.error(`Promise rejection: ${error}`);
    if (error.stack)
        logger.error(error.stack);
});
class PolymerCli {
    constructor(args) {
        this.commandDescriptors = [];
        this.commands = new Map();
        console.log(args);
        // If the "--quiet"/"-q" flag is ever present, set our global logging
        // to quiet mode. Also set the level on the logger we've already created.
        if (args.indexOf('--quiet') > -1 || args.indexOf('-q') > -1) {
            logging.setQuiet();
        }
        // If the "--verbose"/"-v" flag is ever present, set our global logging
        // to verbose mode. Also set the level on the logger we've already created.
        if (args.indexOf('--verbose') > -1 || args.indexOf('-v') > -1) {
            logging.setVerbose();
        }
        logger.debug('got args:', { args: args });
        this.args = args;
        this.addCommand(new build_1.BuildCommand());
        this.addCommand(new help_1.HelpCommand(this.commands));
        this.addCommand(new init_1.InitCommand());
        this.addCommand(new lint_1.LintCommand());
        this.addCommand(new serve_1.ServeCommand());
        this.addCommand(new test_1.TestCommand());
    }
    addCommand(command) {
        logger.debug('adding command', command.name);
        this.commands.set(command.name, command);
        this.commandDescriptors.push({
            name: command.name,
            definitions: this.mergeDefinitions(command, args_1.globalArguments),
            description: command.description,
        });
    }
    mergeDefinitions(command, globals) {
        let mergedArgs = new Map();
        let defaultOption = null;
        let addAll = (args) => {
            for (let definition of args) {
                let name = definition.name;
                let oldDefinition = mergedArgs.get(name);
                if (oldDefinition == null) {
                    mergedArgs.set(definition.name, definition);
                }
                else {
                    let mergedDefinition = Object.assign({}, oldDefinition);
                    for (let propName of Object.keys(definition)) {
                        if (propName == 'name')
                            continue;
                        let propValue = definition[propName];
                        let oldProp = oldDefinition[propName];
                        if (oldProp == null) {
                            mergedDefinition[propName] = propValue;
                        }
                        else {
                            throw new Error(`duplicate argument definition in ${command.name}: ${name}.${propName}`);
                        }
                    }
                    mergedArgs.set(name, mergedDefinition);
                    definition = mergedDefinition;
                }
                if (definition.defaultOption) {
                    if (defaultOption && defaultOption != name) {
                        throw new Error(`Multiple default arguments in ${command.name}: ` +
                            `${defaultOption} and ${name}`);
                    }
                    defaultOption = name;
                }
            }
        };
        if (globals)
            addAll(globals);
        if (command.args)
            addAll(command.args);
        return Array.from(mergedArgs.values());
    }
    run() {
        logger.debug('running...');
        // If the "--version" flag is ever present, just print
        // the current version. Useful for globally installed CLIs.
        if (this.args.indexOf('--version') > -1) {
            console.log(require('../package.json').version);
            return;
        }
        // Alias the "--help"/"-h" flag to run the help command for the given
        // command. This gets us around some limitations in our arguments
        // parsers and keeps us from having to implement additional help logic
        // inside of each command.
        if (this.args.indexOf('--help') > -1 || this.args.indexOf('-h') > -1) {
            this.args = ['help', this.args[0]];
        }
        this.cli = commandLineCommands(this.commandDescriptors);
        let cliCommand = this.cli.parse(this.args);
        logger.debug('command parsed', cliCommand);
        let command = this.commands.get(cliCommand.name || 'help');
        let options = (cliCommand.options
            && cliCommand.options['_all']);
        let config = new project_config_1.ProjectConfig('polymer.json', options);
        logger.debug('config read', config);
        logger.debug('Running command...');
        command.run(options, config).catch((error) => {
            console.error('error', error);
            if (error.stack)
                console.error(error.stack);
        });
    }
}
exports.PolymerCli = PolymerCli;